// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Nathan van Beelen
// Copyright (C) 2019  Mathieu Velten (FLA signed)
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

library matrix_sdk_sqflite;

import 'dart:convert';

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:meta/meta.dart';
import 'package:sqflite/sqflite.dart';
import 'package:url/url.dart';

class SqfliteStore implements Store {
  final String path;

  SqfliteStore({@required this.path});

  bool _isOpen = false;

  @override
  bool get isOpen => _isOpen;

  LocalUser _user;

  Database _db;

  static const sent = 'sent';
  static const unsent = 'unsent';

  static const _localUserTable = 'local_user';
  static const _roomTable = 'room';
  static const _eventTable = 'event';
  static const _timelineTable = 'timeline';
  static const _userTable = 'user';
  static const _memberTable = 'member';
  static const _userStateTable = 'user_state';
  static const _deviceTable = 'device';
  static const _customDataTable = 'custom_data';

  static const _currentUserStateView = 'current_user_state';
  static const _currentAliasView = 'current_alias';

  static const _idColumn = 'id';

  static const _nameColumn = 'name';
  static const _userIdColumn = 'user_id';
  static const _avatarUrlColumn = 'avatar_url';

  static const _accessTokenColumn = 'access_token';
  static const _syncTokenColumn = 'sync_token';
  static const _homeserverColumn = 'homeserver';
  static const _deviceIdColumn = 'device_id';

  static const _roomIdColumn = 'room_id';

  static const _countColumn = 'count';

  // Room columns
  static const _typeColumn = 'type';
  static const _memberCountColumn = 'member_count';
  static const _directUserIdColumn = 'direct_user_id';
  static const _previousBatchColumn = 'previous_batch';
  static const _highlightedUnreadNotificationCountColumn =
      'highlighted_unread_notification_count';
  static const _totalUnreadNotificationCountColumn =
      'total_unread_notification_count';

  static const _nameChangeEventIdColumn = 'name_change_event_id';
  static const _avatarChangeEventIdColumn = 'avatar_change_event_id';
  static const _topicChangeEventIdColumn = 'topic_change_event_id';
  static const _powerLevelsChangeEventIdColumn = 'power_levels_change_event_id';
  static const _creationEventIdColumn = 'creation_event_id';
  static const _upgradeEventIdColumn = 'upgrade_event_id';

  // Event columns
  static const _stateKeyColumn = 'state_key';
  static const _contentColumn = 'content';
  static const _senderIdColumn = 'sender_id';
  static const _timeColumn = 'time';
  static const _sentStateColumn = 'sent_state';
  static const _senderStateSinceColumn = 'sender_state_since';

  // Timeline columns
  static const _eventIdColumn = 'event_id';

  // UserState columns
  static const _sinceColumn = 'since';
  static const _displayNameColumn = 'display_name';

  // Device columns
  static const _lastSeenColumn = 'last_seen';
  static const _lastIpAddressColumn = 'last_ip_address';

  // Custom data columns
  static const _keyColumn = 'key';
  static const _dataColumn = 'data';
  static const _objectIdColumn = 'object_id';

  Future<void> _createTables(Database db, LocalUser user) async {
    if (user == null) {
      throw ArgumentError('user cannot be null if the database does not exist');
    }

    _user = user;

    const tables = [
      '''
      CREATE TABLE IF NOT EXISTS $_localUserTable (
        $_idColumn INTEGER PRIMARY KEY,
        $_userIdColumn TEXT,
        $_displayNameColumn TEXT,
        $_avatarUrlColumn TEXT,
        $_accessTokenColumn TEXT,
        $_syncTokenColumn TEXT,
        $_homeserverColumn TEXT,
        $_deviceIdColumn TEXT,
        FOREIGN KEY ($_userIdColumn) REFERENCES $_userTable($_idColumn),
        FOREIGN KEY ($_deviceIdColumn) REFERENCES $_deviceTable($_idColumn));
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_deviceTable (
        $_idColumn TEXT PRIMARY KEY,
        $_userIdColumn TEXT,
        $_nameColumn TEXT,
        $_lastSeenColumn INT,
        $_lastIpAddressColumn TEXT,
        FOREIGN KEY ($_userIdColumn) REFERENCES $_userTable($_idColumn));
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_roomTable (
        $_idColumn TEXT PRIMARY KEY,
        $_typeColumn TEXT,
        $_memberCountColumn INT NULL,
        $_directUserIdColumn TEXT NULL,
        $_previousBatchColumn TEXT NOT NULL,
        $_highlightedUnreadNotificationCountColumn INT,
        $_totalUnreadNotificationCountColumn INT,
        $_nameChangeEventIdColumn TEXT NULL,
        $_avatarChangeEventIdColumn TEXT NULL,
        $_topicChangeEventIdColumn TEXT NULL,
        $_powerLevelsChangeEventIdColumn TEXT NULL,
        $_creationEventIdColumn TEXT NULL,
        $_upgradeEventIdColumn TEXT NULL,
        FOREIGN KEY ($_directUserIdColumn) REFERENCES $_userTable($_idColumn),
        FOREIGN KEY ($_nameChangeEventIdColumn)
          REFERENCES $_eventTable($_idColumn),
        FOREIGN KEY ($_avatarChangeEventIdColumn)
          REFERENCES $_eventTable($_idColumn),
        FOREIGN KEY ($_avatarChangeEventIdColumn)
          REFERENCES $_eventTable($_idColumn),
        FOREIGN KEY ($_topicChangeEventIdColumn)
          REFERENCES $_eventTable($_idColumn),
        FOREIGN KEY ($_powerLevelsChangeEventIdColumn)
          REFERENCES $_eventTable($_idColumn),
        FOREIGN KEY ($_creationEventIdColumn)
          REFERENCES $_eventTable($_idColumn),
        FOREIGN KEY ($_upgradeEventIdColumn)
          REFERENCES $_eventTable($_idColumn)
        );
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_eventTable (
        $_idColumn TEXT PRIMARY KEY,
        $_typeColumn TEXT NON NULL,
        $_roomIdColumn TEXT,
        $_senderIdColumn TEXT NULL,
        $_senderStateSinceColumn TEXT NOT NULL,
        $_timeColumn INT NOT NULL,
        $_contentColumn TEXT NOT NULL,
        $_sentStateColumn TEXT NOT NULL,
        $_stateKeyColumn TEXT NULL,
        FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn),
        FOREIGN KEY($_senderIdColumn, $_roomIdColumn, $_senderStateSinceColumn)
          REFERENCES $_userStateTable($_userIdColumn,
                                      $_roomIdColumn,
                                      $_sinceColumn)
      );
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_timelineTable (
        $_eventIdColumn TEXT PRIMARY KEY,
        $_roomIdColumn TEXT NON NULL,
        FOREIGN KEY($_eventIdColumn) REFERENCES $_eventTable($_idColumn),
        FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn)
      );
      ''',
      '''
      CREATE VIEW IF NOT EXISTS $_currentAliasView AS
        SELECT $_roomIdColumn,
               $_stateKeyColumn,
               $_contentColumn,
               MAX($_timeColumn) AS $_timeColumn
        FROM $_eventTable
        WHERE type='m.room.aliases'
        GROUP BY $_roomIdColumn, $_stateKeyColumn
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_userTable (
        $_idColumn TEXT PRIMARY KEY);
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_userStateTable (
        $_userIdColumn TEXT,
        $_roomIdColumn TEXT,
        $_sinceColumn INT,
        $_displayNameColumn TEXT NULL,
        $_avatarUrlColumn TEXT NULL,
        FOREIGN KEY($_userIdColumn) REFERENCES $_userTable($_idColumn),
        FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn)
        PRIMARY KEY ($_userIdColumn, $_roomIdColumn, $_sinceColumn));
      ''',
      '''
      CREATE VIEW IF NOT EXISTS $_currentUserStateView AS
        SELECT $_userIdColumn,
               $_roomIdColumn,
               $_displayNameColumn,
               $_avatarUrlColumn,
               MAX($_sinceColumn) AS $_sinceColumn
        FROM $_userStateTable
        GROUP BY $_userIdColumn, $_roomIdColumn
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_memberTable (
        $_userIdColumn TEXT,
        $_roomIdColumn TEXT,
        FOREIGN KEY($_userIdColumn) REFERENCES $_userTable($_idColumn),
        FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn)
        PRIMARY KEY ($_userIdColumn, $_roomIdColumn));
      ''',
      '''
      CREATE TABLE IF NOT EXISTS $_customDataTable (
        $_objectIdColumn TEXT NULL,
        $_keyColumn,
        $_dataColumn TEXT NULL,
        PRIMARY KEY ($_objectIdColumn, $_keyColumn));
        );
    ''',
    ];

    final batch = db.batch();
    tables.forEach(batch.execute);

    await batch.commit(noResult: true);
  }

  @override
  Future<bool> open({LocalUser user}) async {
    try {
      _db = await openDatabase(
        path,
        version: 17,
        onUpgrade: (Database db, int oldVersion, int newVersion) async {
          final localUser = await getLocalUser(db: db);

          // Drop all table and clear sync token
          final batch = db.batch()
            ..execute('UPDATE $_localUserTable SET $_syncTokenColumn=NULL')
            ..execute('DROP TABLE IF EXISTS $_roomTable')
            ..execute('DROP TABLE IF EXISTS $_eventTable')
            ..execute('DROP TABLE IF EXISTS $_userTable')
            ..execute('DROP TABLE IF EXISTS $_memberTable')
            ..execute('DROP TABLE IF EXISTS $_userStateTable')
            ..execute('DROP TABLE IF EXISTS $_deviceTable')
            ..execute('DROP VIEW  IF EXISTS $_currentUserStateView');

          await batch.commit(noResult: true);

          await _createTables(db, localUser);
        },
        onCreate: (Database db, int version) async {
          await _createTables(db, user);
          await setLocalUser(user, db: db);
        },
      );

      if (_user == null) {
        _user = await getLocalUser(db: _db);
      }
    } on ArgumentError {
      return false;
    }

    _isOpen = true;
    return true;
  }

  @override
  Future<void> close() async {
    _isOpen = false;
    await _db.close();
  }

  @override
  Future<void> delete() async {
    await _db?.close();
    await deleteDatabase(await getDatabasesPath() + path);
  }

  @override
  Future<LocalUser> getLocalUser({Database db}) async {
    if (db == null) {
      db = _db;
    }

    const deviceName = 'device_name';

    final results = await db.rawQuery('''
        SELECT $_localUserTable.$_userIdColumn,
               $_localUserTable.$_displayNameColumn,
               $_localUserTable.$_avatarUrlColumn,
               $_localUserTable.$_accessTokenColumn,
               $_localUserTable.$_homeserverColumn,
               $_localUserTable.$_deviceIdColumn,
               $_deviceTable.$_nameColumn AS $deviceName,
               $_deviceTable.$_lastSeenColumn,
               $_deviceTable.$_lastIpAddressColumn
        FROM $_localUserTable
        LEFT JOIN $_deviceTable
          ON $_deviceTable.$_idColumn=$_localUserTable.$_deviceIdColumn
        WHERE $_localUserTable.$_idColumn = 1
      ''');

    if (results.isNotEmpty) {
      final result = results.first;

      final userId = UserId(result[_userIdColumn]);

      var lastSeen = result[_lastSeenColumn];
      if (lastSeen != null) {
        lastSeen = DateTime.fromMillisecondsSinceEpoch(lastSeen);
      }

      var avatarUrl = result[_avatarUrlColumn];
      if (avatarUrl != null) {
        avatarUrl = Url.parse(avatarUrl);
      }

      return LocalUser(
        id: userId,
        name: result[_displayNameColumn],
        avatarUrl: avatarUrl,
        accessToken: result[_accessTokenColumn],
        homeserver: Homeserver(Url.parse(result[_homeserverColumn])),
        currentDevice: Device(
          id: DeviceId(result[_deviceIdColumn]),
          userId: userId,
          name: result[deviceName],
          lastSeen: lastSeen,
          lastIpAddress: result[_lastIpAddressColumn],
        ),
        store: this,
      );
    } else {
      return null;
    }
  }

  @override
  Future<void> setLocalUser(LocalUser user, {Database db}) async {
    if (db == null) {
      db = _db;
    }

    await db.transaction((txn) async {
      await txn.execute(
        '''
        INSERT OR REPLACE INTO $_localUserTable($_idColumn,
                                                $_userIdColumn,
                                                $_displayNameColumn,
                                                $_avatarUrlColumn,
                                                $_accessTokenColumn,
                                                $_homeserverColumn,
                                                $_deviceIdColumn)
          VALUES (?, ?, ?, ?, ?, ?, ?);
        ''',
        [
          1,
          user.id.toString(),
          user.name,
          user.avatarUrl?.toString(),
          user.accessToken,
          user.homeserver.url.toString(),
          user.currentDevice.id.toString(),
        ],
      );

      await txn.execute('''
        INSERT OR REPLACE INTO $_deviceTable($_idColumn,
                                             $_userIdColumn,
                                             $_nameColumn,
                                             $_lastSeenColumn,
                                             $_lastIpAddressColumn)
         VALUES (?, ?, ?, ?, ?);
        ''', [
        user.currentDevice.id.toString(),
        user.currentDevice.userId.toString(),
        user.currentDevice.name,
        user.currentDevice.lastSeen?.millisecondsSinceEpoch,
        user.currentDevice.lastIpAddress,
      ]);
    });
  }

  @override
  Future<String> getSyncToken() async {
    final results = await _db.rawQuery(
      'SELECT $_syncTokenColumn FROM $_localUserTable WHERE $_idColumn = 1;',
    );

    if (results.isNotEmpty) {
      return results.first[_syncTokenColumn];
    } else {
      return null;
    }
  }

  @override
  Future<void> setSyncToken(String syncToken) async {
    await _db.transaction((txn) async {
      await txn.execute(
        '''
        UPDATE $_localUserTable SET $_syncTokenColumn=?
        ''',
        [syncToken],
      );
    });
  }

  static const _userAvatarUrlColumn = 'user_avatar_url';

  static const _nameChangeJoin = 'name_change';
  static const _avatarChangeJoin = 'avatar_change';
  static const _topicChangeJoin = 'topic_change';
  static const _powerLevelsChangeJoin = 'power_levels_change';
  static const _creationJoin = 'creation';
  static const _upgradeJoin = 'upgrade';

  static const _nameChangeContentColumn = 'name_change_content';
  static const _nameChangeSentStateColumn = 'name_change_sent_state';
  static const _nameChangeSenderIdColumn = 'name_change_sender_id';

  static const _avatarChangeContentColumn = 'avatar_change_content';
  static const _avatarChangeSentStateColumn = 'avatar_change_sent_state';
  static const _avatarChangeSenderIdColumn = 'avatar_change_sender_id';

  static const _topicChangeContentColumn = 'topic_change_content';
  static const _topicChangeSentStateColumn = 'topic_change_sent_state';
  static const _topicChangeSenderIdColumn = 'topic_change_sender_id';

  static const _powerLevelsChangeContentColumn = 'power_levels_change_content';
  static const _powerLevelsChangeSentStateColumn =
      'power_levels_change_sent_state';
  static const _powerLevelsChangeSenderIdColumn =
      'power_levels_change_sender_id';

  static const _creationContentColumn = 'creation_change_content';
  static const _creationSentStateColumn = 'creation_change_sent_state';
  static const _creationSenderIdColumn = 'creation_change_sender_id';

  static const _upgradeContentColumn = 'upgrade_change_content';
  static const _upgradeSentStateColumn = 'upgrade_change_sent_state';
  static const _upgradeSenderIdColumn = 'upgrade_change_sender_id';

  static const _roomSelectBase = '''
    SELECT $_roomTable.rowid,
           $_roomTable.$_idColumn,
           $_roomTable.$_typeColumn,
           $_roomTable.$_memberCountColumn,
           $_roomTable.$_directUserIdColumn,
           $_roomTable.$_previousBatchColumn,
           $_roomTable.$_highlightedUnreadNotificationCountColumn,
           $_roomTable.$_totalUnreadNotificationCountColumn,
           $_currentUserStateView.$_displayNameColumn,
           $_currentUserStateView.$_avatarUrlColumn AS $_userAvatarUrlColumn,
           $_currentUserStateView.$_sinceColumn,
           $_nameChangeJoin.$_contentColumn AS $_nameChangeContentColumn,
           $_nameChangeJoin.$_sentStateColumn AS $_nameChangeSentStateColumn,
           $_nameChangeJoin.$_senderIdColumn AS $_nameChangeSenderIdColumn,
           $_avatarChangeJoin.$_contentColumn AS $_avatarChangeContentColumn,
           $_avatarChangeJoin.$_sentStateColumn AS $_avatarChangeSentStateColumn,
           $_avatarChangeJoin.$_senderIdColumn AS $_avatarChangeSenderIdColumn,
           $_topicChangeJoin.$_contentColumn AS $_topicChangeContentColumn,
           $_topicChangeJoin.$_sentStateColumn AS $_topicChangeSentStateColumn,
           $_topicChangeJoin.$_senderIdColumn AS $_topicChangeSenderIdColumn,
           $_powerLevelsChangeJoin.$_contentColumn AS $_powerLevelsChangeContentColumn,
           $_powerLevelsChangeJoin.$_sentStateColumn AS $_powerLevelsChangeSentStateColumn,
           $_powerLevelsChangeJoin.$_senderIdColumn AS $_powerLevelsChangeSenderIdColumn,
           $_creationJoin.$_contentColumn AS $_creationContentColumn,
           $_creationJoin.$_sentStateColumn AS $_creationSentStateColumn,
           $_creationJoin.$_senderIdColumn AS $_creationSenderIdColumn,
           $_upgradeJoin.$_contentColumn AS $_upgradeContentColumn,
           $_upgradeJoin.$_sentStateColumn AS $_upgradeSentStateColumn,
           $_upgradeJoin.$_senderIdColumn AS $_upgradeSenderIdColumn
    FROM $_roomTable
    LEFT JOIN $_currentUserStateView
      ON $_currentUserStateView.$_userIdColumn=$_roomTable.$_directUserIdColumn
       AND $_currentUserStateView.$_roomIdColumn=$_roomTable.$_idColumn
    LEFT JOIN $_eventTable $_nameChangeJoin
      ON $_nameChangeJoin.$_idColumn=$_roomTable.$_nameChangeEventIdColumn
    LEFT JOIN $_eventTable $_avatarChangeJoin
      ON $_avatarChangeJoin.$_idColumn=$_roomTable.$_avatarChangeEventIdColumn
    LEFT JOIN $_eventTable $_topicChangeJoin
      ON $_topicChangeJoin.$_idColumn=$_roomTable.$_topicChangeEventIdColumn
    LEFT JOIN $_eventTable $_powerLevelsChangeJoin
      ON $_powerLevelsChangeJoin.$_idColumn=$_roomTable.$_powerLevelsChangeEventIdColumn
    LEFT JOIN $_eventTable $_creationJoin
      ON $_creationJoin.$_idColumn=$_roomTable.$_creationEventIdColumn
    LEFT JOIN $_eventTable $_upgradeJoin
      ON $_upgradeJoin.$_idColumn=$_roomTable.$_upgradeEventIdColumn
    ''';

  Future<Room> _parseRoomRow(Map<String, dynamic> row,
      {Transaction txn}) async {
    final id = RoomId(row[_idColumn]);
    final type = row[_typeColumn];

    final memberCount = row[_memberCountColumn];

    final previousBatch = row[_previousBatchColumn];

    final directUserId = row[_directUserIdColumn];
    User directUser;
    if (directUserId != null) {
      final displayName = row[_displayNameColumn];

      var userAvatarUrl = row[_userAvatarUrlColumn];
      if (userAvatarUrl != null) {
        userAvatarUrl = Url.parse(userAvatarUrl);
      }

      var since = row[_sinceColumn];
      if (since != null) {
        since = DateTime.fromMillisecondsSinceEpoch(since);
      }

      directUser = User(
        id: UserId(directUserId),
        state: UserState(
          roomId: id,
          since: since,
          displayName: displayName,
          avatarUrl: userAvatarUrl,
        ),
      );
    }

    final hightlightedNotifCount =
        row[_highlightedUnreadNotificationCountColumn] ?? 0;
    final totalNotifCount = row[_totalUnreadNotificationCountColumn] ?? 0;

    var nameChangeEvent = row[_nameChangeContentColumn];
    if (nameChangeEvent != null) {
      nameChangeEvent = await RoomEvent.fromJson(
        RoomEventArgs(roomId: id),
        json.decode(nameChangeEvent),
      );
    }

    var avatarChangeEvent = row[_avatarChangeContentColumn];
    if (avatarChangeEvent != null) {
      avatarChangeEvent = await RoomEvent.fromJson(
        RoomEventArgs(roomId: id),
        json.decode(avatarChangeEvent),
      );
    }

    var topicChangeEvent = row[_topicChangeContentColumn];
    if (topicChangeEvent != null) {
      topicChangeEvent = await RoomEvent.fromJson(
        RoomEventArgs(roomId: id),
        json.decode(topicChangeEvent),
      );
    }

    var powerLevelsChangeEvent = row[_powerLevelsChangeContentColumn];
    if (powerLevelsChangeEvent != null) {
      powerLevelsChangeEvent = await RoomEvent.fromJson(
        RoomEventArgs(roomId: id),
        json.decode(powerLevelsChangeEvent),
      );
    }

    var creationEvent = row[_creationContentColumn];
    if (creationEvent != null) {
      creationEvent = await RoomEvent.fromJson(
        RoomEventArgs(roomId: id),
        json.decode(creationEvent),
      );
    }

    var upgradeEvent = row[_upgradeContentColumn];
    if (upgradeEvent != null) {
      upgradeEvent = await RoomEvent.fromJson(
        RoomEventArgs(roomId: id),
        json.decode(upgradeEvent),
      );
    }

    Map<String, AliasesChangeEvent> aliasChanges = Map();
    // TODO: Optimize
    Future<void> setAliases(Transaction txn) async {
      final results = await txn.rawQuery(
        '''
        SELECT $_contentColumn FROM $_currentAliasView WHERE $_roomIdColumn=?
        ''',
        [id.toString()],
      );

      for (final result in results) {
        final event = await RoomEvent.fromJson(
          RoomEventArgs(roomId: id),
          json.decode(result[_contentColumn]),
        ) as AliasesChangeEvent;

        aliasChanges[event.stateKey] = event;
      }
    }

    if (txn == null) {
      await _db.transaction(setAliases);
    } else {
      await setAliases(txn);
    }

    final stateEvents = RoomStateEvents(
      nameChange: nameChangeEvent,
      avatarChange: avatarChangeEvent,
      topicChange: topicChangeEvent,
      powerLevelsChange: powerLevelsChangeEvent,
      aliasChanges: aliasChanges,
      creation: creationEvent,
      upgrade: upgradeEvent,
    );

    if (type == 'join') {
      return JoinedRoom(
        id: id,
        memberCount: memberCount,
        previousBatch: previousBatch,
        scope: _user.scope,
        directUser: directUser,
        highlightedUnreadNotificationCount: hightlightedNotifCount,
        totalUnreadNotificationCount: totalNotifCount,
        stateEvents: stateEvents,
      );
    } else {
      return LeftRoom(
        id: id,
        memberCount: memberCount,
        previousBatch: previousBatch,
        scope: _user.scope,
        directUser: directUser,
        highlightedUnreadNotificationCount: hightlightedNotifCount,
        totalUnreadNotificationCount: totalNotifCount,
        stateEvents: stateEvents,
      );
    }
  }

  @override
  Future<Room> getRoom(RoomId id) async {
    final results = await _db.rawQuery(
      _roomSelectBase + ' WHERE $_roomTable.$_idColumn = ?;',
      [id.toString()],
    );

    if (results.isNotEmpty) {
      return _parseRoomRow(results.first);
    } else {
      return null;
    }
  }

  @override
  Future<Iterable<Room>> getRooms({Room from, int count}) async {
    var query = _roomSelectBase;

    var values = <dynamic>[];

    if (from != null) {
      query += ' WHERE $_roomTable.rowid > '
          '(SELECT $_roomTable.rowid FROM $_roomTable'
          ' WHERE $_roomTable.$_idColumn = ?)';
      values.add(from.id.toString());
    }

    query += ' ORDER BY $_roomTable.$_idColumn';

    if (count != null) {
      query += ' LIMIT ?';
      values.add(count);
    }

    var results = await _db.rawQuery(query, values);

    return await Future.wait(results.map((row) => _parseRoomRow(row)));
  }

  @override
  Future<int> getRoomCount() async {
    final results = await _db.rawQuery(
        'SELECT COUNT($_idColumn) as $_countColumn FROM $_roomTable;');

    if (results.isNotEmpty) {
      return results.first[_countColumn];
    } else {
      return -1;
    }
  }

  @override
  Future<void> setRooms(Iterable<Room> rooms) async {
    await _db.transaction((txn) async {
      for (Room room in rooms) {
        var type;
        if (room is JoinedRoom) {
          type = 'join';
        } else if (room is LeftRoom) {
          type = 'leave';
        }

        await txn.execute(
          '''
          INSERT OR REPLACE INTO $_roomTable(
            $_idColumn,
            $_typeColumn,
            $_memberCountColumn,
            $_directUserIdColumn,
            $_previousBatchColumn,
            $_highlightedUnreadNotificationCountColumn,
            $_totalUnreadNotificationCountColumn,
            $_nameChangeEventIdColumn,
            $_avatarChangeEventIdColumn,
            $_topicChangeEventIdColumn,
            $_powerLevelsChangeEventIdColumn,
            $_creationEventIdColumn,
            $_upgradeEventIdColumn)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
          ''',
          [
            room.id.toString(),
            type,
            room.members.count,
            room.directUser?.id?.toString(),
            room.previousBatch,
            room.highlightedUnreadNotificationCount,
            room.totalUnreadNotificationCount,
            room.stateEvents.nameChange?.id?.toString(),
            room.stateEvents.avatarChange?.id?.toString(),
            room.stateEvents.topicChange?.id?.toString(),
            room.stateEvents.powerLevelsChange?.id?.toString(),
            room.stateEvents.creation?.id?.toString(),
            room.stateEvents.upgrade?.id?.toString(),
          ],
        );

        await addRoomEvents(
          room,
          [
            room.stateEvents.nameChange,
            room.stateEvents.avatarChange,
            room.stateEvents.topicChange,
            room.stateEvents.powerLevelsChange,
            room.stateEvents.creation,
            room.stateEvents.upgrade,
          ]
              .where((e) => e != null)
              .followedBy(room.stateEvents.aliasChanges.values),
          inTimeline: false,
          txn: txn,
        );
      }
    });
  }

  @override
  Future<void> removeEvent(EventId id) async {
    await _db.rawQuery(
      '''
      DELETE FROM $_eventTable WHERE $_idColumn = ?;
      ''',
      [id.toString()],
    );
  }

  static const _roomEventSelectBase = '''
    SELECT $_eventTable.$_contentColumn,
           $_eventTable.$_sentStateColumn,
           $_eventTable.$_senderIdColumn,
           $_userStateTable.$_sinceColumn,
           $_userStateTable.$_displayNameColumn,
           $_userStateTable.$_avatarUrlColumn
    FROM $_eventTable
    INNER JOIN $_timelineTable
      ON $_timelineTable.$_eventIdColumn=$_eventTable.$_idColumn
    LEFT OUTER JOIN $_userStateTable
      ON $_userStateTable.$_userIdColumn=$_eventTable.$_senderIdColumn
        AND $_userStateTable.$_roomIdColumn=$_eventTable.$_roomIdColumn
        AND $_userStateTable.$_sinceColumn=$_eventTable.$_senderStateSinceColumn
    WHERE $_eventTable.$_roomIdColumn = ?
    ''';

  Future<RoomEvent> _parseRoomEventRow(
    Room room,
    Map<String, dynamic> row, {
    ProcessContext context,
  }) async {
    final content = row[_contentColumn];
    final sentState =
        row[_sentStateColumn] == 'sent' ? SentState.sent : SentState.unsent;

    final displayName = row[_displayNameColumn];
    var avatarUrl = row[_avatarUrlColumn];
    if (avatarUrl != null) {
      avatarUrl = Url.parse(avatarUrl);
    }

    var since = row[_sinceColumn];
    if (since != null) {
      since = DateTime.fromMillisecondsSinceEpoch(since);
    }

    final user = User(
      id: UserId(row[_senderIdColumn]),
      state: UserState(
        roomId: room.id,
        since: since,
        displayName: displayName,
        avatarUrl: avatarUrl,
      ),
    );

    return await RoomEvent.fromJson(
      RoomEventArgs(roomId: room.id, sender: user, sentState: sentState),
      json.decode(content),
      context: context,
    );
  }

  @override
  Future<RoomEvent> getRoomEvent(Room room, EventId id) async {
    var results = await _db.rawQuery(
      _roomEventSelectBase + ' AND $_eventTable.$_idColumn = ?',
      [room.id.toString(), id.toString()],
    );

    if (results.isNotEmpty) {
      return _parseRoomEventRow(room, results.first);
    } else {
      return null;
    }
  }

  @override
  Future<Iterable<RoomEvent>> getRoomEvents(
    Room room, {
    RoomEvent from,
    int count,
  }) async {
    final context = ProcessContext(members: room.members);

    var query = _roomEventSelectBase;

    var values = <dynamic>[room.id.toString()];

    if (from != null) {
      query += ' AND $_eventTable.$_timeColumn < ?';
      values.add(from.time.millisecondsSinceEpoch);
    }

    query += ' ORDER BY $_eventTable.$_timeColumn DESC';

    if (count != null) {
      query += ' LIMIT ?';
      values.add(count);
    }

    var results = await _db.rawQuery(query, values);

    final events = List<RoomEvent>();
    for (final result in results) {
      events.add(await _parseRoomEventRow(room, result, context: context));
    }

    return events;
  }

  @override
  Future<int> getRoomEventCount(RoomId roomId) async {
    final results = await _db.rawQuery(
      '''
    SELECT COUNT($_eventTable.$_idColumn) as $_countColumn FROM $_eventTable
      WHERE $_eventTable.$_roomIdColumn = ?;
    ''',
      [roomId.toString()],
    );

    if (results.isNotEmpty) {
      return results.first[_countColumn];
    } else {
      return -1;
    }
  }

  @override
  Future<Iterable<RoomEvent>> getUnsentEvents(Room room) async {
    final results = await _db.rawQuery(
      '''
      SELECT $_eventTable.$_contentColumn
      FROM $_eventTable
      WHERE $_eventTable.$_roomIdColumn = ?
        AND $_eventTable.$_sentStateColumn = ?
      ''',
      [room.id.toString(), unsent],
    );

    if (results.isNotEmpty) {
      return Future.wait(results.map((result) async {
        final content = result[_contentColumn];

        return await RoomEvent.fromJson(
          RoomEventArgs(
            roomId: room.id,
            sender: _user,
            sentState: SentState.unsent,
          ),
          json.decode(content),
        );
      }));
    } else {
      return Iterable.empty();
    }
  }

  @override
  Future<void> addRoomEvents(
    Room room,
    Iterable<RoomEvent> events, {
    bool inTimeline = true,
    Transaction txn,
  }) async {
    Future<void> process(Transaction txn) async {
      for (Event event in events) {
        var senderId;
        var senderStateSince;

        var time;
        final sentState = event.sentState != SentState.sent ? unsent : sent;

        if (event is RoomEvent) {
          senderId = event.sender.id.toString();
          time = event.time.millisecondsSinceEpoch;
          senderStateSince = event.sender.state?.since?.millisecondsSinceEpoch;
        }

        var stateKey;
        if (event is StateEvent) {
          stateKey = event.stateKey;
        }

        await txn.execute(
          '''
          INSERT OR IGNORE INTO $_eventTable($_idColumn,
                                             $_typeColumn,
                                             $_roomIdColumn,
                                             $_timeColumn,
                                             $_senderIdColumn,
                                             $_senderStateSinceColumn,
                                             $_contentColumn,
                                             $_sentStateColumn,
                                             $_stateKeyColumn)
            values (?, ?, ?, ?, ?, ?, ?, ?, ?);
          ''',
          [
            event.id.toString(),
            Event.matrixTypeOf(event),
            room.id.toString(),
            time,
            senderId,
            senderStateSince,
            json.encode(event.toJson()),
            sentState,
            stateKey,
          ],
        );

        if (inTimeline) {
          await txn.execute(
            '''
          INSERT OR IGNORE INTO $_timelineTable($_eventIdColumn, $_roomIdColumn)
            VALUES (?, ?);
          ''',
            [
              event.id.toString(),
              room.id.toString(),
            ],
          );
        }
      }
    }

    if (txn == null) {
      await _db.transaction(process);
    } else {
      await process(txn);
    }
  }

  Future<void> _setUser(Transaction txn, User user) async {
    var url;
    if (user.avatarUrl != null) {
      url = user.avatarUrl.toString();
    }

    await txn.execute(
      '''
      INSERT OR IGNORE INTO $_userTable($_idColumn)
        VALUES (?);
      ''',
      [user.id.toString()],
    );

    if (user.state != null) {
      await txn.execute(
        '''
      INSERT OR IGNORE INTO $_userStateTable($_userIdColumn,
                                             $_roomIdColumn,
                                             $_sinceColumn,
                                             $_displayNameColumn,
                                             $_avatarUrlColumn)
        VALUES (?, ?, ?, ?, ?);
      ''',
        [
          user.id.toString(),
          user.state.roomId.toString(),
          user.state.since.millisecondsSinceEpoch,
          user.state.displayName,
          url,
        ],
      );
    }
  }

  static const _memberSelectBase = '''
    SELECT $_memberTable.$_userIdColumn,
           $_currentUserStateView.$_roomIdColumn,
           $_currentUserStateView.$_displayNameColumn,
           $_currentUserStateView.$_avatarUrlColumn,
           $_currentUserStateView.$_sinceColumn
    FROM $_memberTable
      LEFT JOIN $_currentUserStateView
        ON
         $_currentUserStateView.$_userIdColumn=$_memberTable.$_userIdColumn
          AND
           $_currentUserStateView.$_roomIdColumn=$_memberTable.$_roomIdColumn
    WHERE $_memberTable.$_roomIdColumn = ?
    ''';

  @override
  Future<User> getMember(RoomId roomId, UserId id, {DateTime at}) async {
    final query = _memberSelectBase + ' AND $_memberTable.$_userIdColumn = ?';
    final values = [roomId.toString(), id.toString()];

    final results = await _db.rawQuery(query, values);

    if (results.isNotEmpty) {
      final first = results.first;
      final user = _parseUserRow(first, id);

      return user;
    } else {
      return null;
    }
  }

  @override
  Future<Iterable<User>> getMembers(
    RoomId roomId, {
    User from,
    int count,
  }) async {
    var query = _memberSelectBase;
    var values = <dynamic>[roomId.toString()];

    if (from != null) {
      query += ' AND $_memberTable.rowid >'
          ' (SELECT $_memberTable.rowid FROM $_memberTable'
          '  WHERE $_memberTable.$_userIdColumn = ?'
          '   AND $_currentUserStateView.$_roomIdColumn = ?'
          '   AND $_currentUserStateView.$_sinceColumn = ?)';
      values.addAll([
        from.id.toString(),
        from.state.roomId.toString(),
        from.state.since.millisecondsSinceEpoch,
      ]);
    }

    query += ' ORDER BY $_currentUserStateView.$_displayNameColumn,'
        ' $_memberTable.$_userIdColumn';

    if (count != null) {
      query += ' LIMIT ?';
      values.add(count);
    }

    final results = await _db.rawQuery(query, values);

    if (results.isNotEmpty) {
      return results.map((result) {
        final id = UserId(result[_userIdColumn]);
        return _parseUserRow(result, id);
      });
    } else {
      return Iterable.empty();
    }
  }

  @override
  Future<int> getMemberCount(RoomId roomId) async {
    final results = await _db.rawQuery(
      '''
      SELECT COUNT($_memberTable.$_userIdColumn) AS $_countColumn
      FROM $_memberTable
        WHERE $_memberTable.$_roomIdColumn = ?;
      ''',
      [roomId.toString()],
    );

    if (results.isNotEmpty) {
      return results.first[_countColumn];
    } else {
      return -1;
    }
  }

  User _parseUserRow(Map<String, dynamic> row, UserId id) {
    var roomId = row[_roomIdColumn];
    if (roomId != null) {
      roomId = RoomId(roomId);
    }

    var since = row[_sinceColumn];
    if (since != null) {
      since = DateTime.fromMicrosecondsSinceEpoch(row[_sinceColumn]);
    }

    final displayName = row[_displayNameColumn];
    var avatarUrl = row[_avatarUrlColumn];
    if (avatarUrl != null) {
      avatarUrl = Url.parse(avatarUrl);
    }

    return User(
      id: id,
      state: UserState(
        roomId: roomId,
        since: since,
        displayName: displayName,
        avatarUrl: avatarUrl,
      ),
    );
  }

  @override
  Future<void> addMembers(RoomId roomId, Iterable<User> members) async {
    await _db.transaction((txn) async {
      for (User member in members) {
        await _setUser(txn, member);

        await txn.execute(
          '''
          INSERT OR IGNORE INTO $_memberTable($_userIdColumn, 
                                              $_roomIdColumn)
            values (?, ?);
          ''',
          [member.id.toString(), roomId.toString()],
        );
      }
    });
  }

  @override
  Future<String> getObjectCustomData(Id objectId, String key) async {
    if (objectId == null) {
      throw Exception();
    }

    final results = await _db.rawQuery(
      '''
      SELECT * FROM $_customDataTable
      WHERE $_customDataTable.$_objectIdColumn = ?
      AND $_customDataTable.$_keyColumn = ?;
      ''',
      [objectId.toString(), key],
    );

    if (results.isNotEmpty) {
      return results[0][_dataColumn];
    }
    return null;
  }

  @override
  Future<void> setObjectCustomData(
    Id objectId,
    String key,
    String value,
  ) async {
    if (objectId == null) {
      throw Exception();
    }
    await _db.transaction((txn) async {
      await txn.execute(
        '''
        INSERT OR REPLACE INTO $_customDataTable($_objectIdColumn,
                                               $_keyColumn,
                                               $_dataColumn)
          values (?, ?, ?);
        ''',
        [objectId.toString(), key, value],
      );
    });
  }

  @override
  Future<String> getAppCustomData(String key) async {
    var results = await _db.rawQuery(
      '''
      SELECT * FROM $_customDataTable
      WHERE $_customDataTable.$_keyColumn = ?
      ''',
      [key],
    );

    if (results.isNotEmpty) {
      return results[0][_dataColumn];
    }

    return null;
  }

  @override
  Future<void> setAppCustomData(String key, String value) async {
    await _db.transaction((txn) async {
      await txn.execute(
        '''
          INSERT OR REPLACE INTO $_customDataTable($_keyColumn,
                                                $_dataColumn)
            values (?, ?);
          ''',
        [key, value],
      );
    });
  }
}
